from typing import Sequence

from sqlalchemy.sql import select

from pypermission import RBAC, Permission
from pypermission.example.models import UserORM, GroupORM, Context, ExampleError, State

################################################################################
#### UserService
################################################################################


class GroupService:

    @classmethod
    def create(
        cls,
        *,
        groupname: str,
        description: str = "A new group.",
        owner: str,
        ctx: Context,
        rbac: bool = True,
    ) -> GroupORM:
        match rbac, owner, ctx.username:
            case True, str(), str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if ctx.username != owner:
                    raise ExampleError(
                        f"User '{ctx.username}' cannot create a group on behalf of '{owner}'."
                    )

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="base",
                        resource_id="",
                        action="group:create",
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case True, str(), None:
                raise ExampleError("No user in context!")
            case False, str(), None:
                ...
            case False, str(), str():
                ...
            case _:
                raise ExampleError("Invalid combination of arguments!")

        group_orm = GroupORM(groupname=groupname, description=description, owner=owner)
        ctx.db.add(group_orm)
        cls._create_group_role_and_policies(groupname=groupname, owner=owner, ctx=ctx)
        return group_orm

    @classmethod
    def list(
        cls,
        *,
        ctx: Context,
        rbac: bool = True,
    ) -> Sequence[GroupORM]:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="group", resource_id="*", action="access"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        return ctx.db.scalars(select(GroupORM)).all()

    @classmethod
    def get(
        cls,
        *,
        groupname: str,
        ctx: Context,
        rbac: bool = True,
    ) -> GroupORM:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="group", resource_id=groupname, action="access"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        group_orm = ctx.db.get(GroupORM, groupname)
        if group_orm is None:
            raise ExampleError(f"Unknown group '{groupname}'!")
        return group_orm

    @classmethod
    def set_description(
        cls,
        *,
        groupname: str,
        description: str,
        ctx: Context,
        rbac: bool = True,
    ) -> GroupORM:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="group", resource_id=groupname, action="edit"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        group_orm = ctx.db.get(GroupORM, groupname)
        if group_orm is None:
            raise ExampleError(f"Unknown group '{groupname}'!")
        group_orm.description = description
        ctx.db.flush()
        return group_orm

    @classmethod
    def set_state(
        cls,
        *,
        groupname: str,
        state: State,
        ctx: Context,
        rbac: bool = True,
    ) -> GroupORM:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="group",
                        resource_id=groupname,
                        action="deactivate",
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        group_orm = ctx.db.get(GroupORM, groupname)
        if group_orm is None:
            raise ExampleError(f"Unknown group '{groupname}'!")
        group_orm.state = state
        ctx.db.flush()
        return group_orm

    @classmethod
    def delete(
        cls,
        *,
        groupname: str,
        ctx: Context,
        rbac: bool = True,
    ) -> GroupORM:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="group", resource_id=groupname, action="delete"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        group_orm = ctx.db.get(GroupORM, groupname)
        if group_orm is None:
            raise ExampleError(f"Unknown group '{groupname}'!")

        GROUP_ROLE__OWNER = f"group[{groupname}]:owner"
        RBAC.role.delete(role=GROUP_ROLE__OWNER, db=ctx.db)

        ctx.db.delete(group_orm)
        ctx.db.flush()
        return group_orm

    ################################################################################
    #### Util
    ################################################################################

    @classmethod
    def _create_group_role_and_policies(
        cls, groupname: str, owner: str, ctx: Context
    ) -> None:
        GROUP_ROLE__OWNER = f"group[{groupname}]_owner"
        RBAC.role.create(role=GROUP_ROLE__OWNER, db=ctx.db)
        RBAC.subject.assign_role(subject=owner, role=GROUP_ROLE__OWNER, db=ctx.db)

        RBAC.role.grant_permission(
            role=GROUP_ROLE__OWNER,
            permission=Permission(
                resource_type="group", resource_id=groupname, action="edit"
            ),
            db=ctx.db,
        )
        RBAC.role.grant_permission(
            role=GROUP_ROLE__OWNER,
            permission=Permission(
                resource_type="group", resource_id=groupname, action="deactivate"
            ),
            db=ctx.db,
        )



"""
`PyPermission` - The python RBAC library for projects where `sqlalchemy` is a valid option.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

__version__ = "0.2.1"
__author__ = "Jendrik A. Potyka, Fabian A. Preiss"

from typing import Final

from pypermission.service.role import RoleService
from pypermission.service.subject import SubjectService
from pypermission.models import (
    Policy,
    Permission,
    FrozenClass,
    create_rbac_database_table,
    set_sqlite_pragma,
)
from pypermission.exc import PyPermissionError, PermissionNotGrantedError


class RBAC(metaclass=FrozenClass):
    """
    Namespace for the Role and Subject services.

    Attributes
    ----------
    role : RoleService
        Shorthand for all RoleService functions.
    subject : SubjectService
        Shorthand for all SubjectService functions.
    """

    role: Final = RoleService
    subject: Final = SubjectService


__all__ = [
    "RBAC",
    "Policy",
    "Permission",
    "create_rbac_database_table",
    "set_sqlite_pragma",
    "PyPermissionError",
    "PermissionNotGrantedError",
]



from sqlalchemy.sql import select
from sqlalchemy.orm import Session
from sqlalchemy.exc import IntegrityError
from collections.abc import Sequence
from pypermission.models import (
    Policy,
    Permission,
    RoleORM,
    HierarchyORM,
    SubjectORM,
    MemberORM,
    PolicyORM,
    FrozenClass,
)
from pypermission.exc import (
    PyPermissionError,
    PermissionNotGrantedError,
    process_subject_role_integrity_error,
)


################################################################################
#### SubjectService
################################################################################


class SubjectService(metaclass=FrozenClass):

    @classmethod
    def create(cls, *, subject: str, db: Session) -> None:
        """
        Create a new Subject.

        Parameters
        ----------
        subject : str
            The SubjectID of the Subject to create.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If a Subject with the given SubjectID already exists.
        """
        try:
            subject_orm = SubjectORM(id=subject)
            db.add(subject_orm)
            db.flush()
        except IntegrityError:
            db.rollback()
            raise PyPermissionError(f"Subject '{subject}' already exists!")

    @classmethod
    def delete(cls, *, subject: str, db: Session) -> None:
        """
        Delete an existing Subject.

        Parameters
        ----------
        subject : str
            The SubjectID to delete.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If a Subject with the given SubjectID does not exist.
        """
        subject_orm = db.get(SubjectORM, subject)
        if subject_orm is None:
            raise PyPermissionError(f"Subject '{subject}' does not exist!")
        db.delete(subject_orm)
        db.flush()

    @classmethod
    def list(cls, *, db: Session) -> tuple[str, ...]:
        """
        Get all Subjects.

        Parameters
        ----------
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[str, ...]
            A tuple containing all SubjectIDs.
        """
        subjects = db.scalars(select(SubjectORM.id)).all()
        return tuple(subjects)

    @classmethod
    def assign_role(cls, *, subject: str, role: str, db: Session) -> None:
        """
        Assign a Subject to a Role.

        Parameters
        ----------
        subject : str
            The target SubjectID.
        role : str
            The target RoleID.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If the Subject does not exist.
            If the Role does not exist.
            If the Subject was assigned to Role before. TODO
        """
        try:
            member_orm = MemberORM(role_id=role, subject_id=subject)
            db.add(member_orm)
            db.flush()
        except IntegrityError as err:
            db.rollback()
            process_subject_role_integrity_error(err=err, subject=subject, role=role)

    @classmethod
    def deassign_role(cls, *, subject: str, role: str, db: Session) -> None:
        """
        Deassign a Subject from a Role.

        Parameters
        ----------
        subject : str
            The target SubjectID.
        role : str
            The target RoleID.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If the Subject does not exist.
            If the Role does not exist.
            If the Subject is not assigned to the Role. TODO
        """
        # TODO raise IntegrityError if subject or role is unknown and if possible via ORM
        member_orm = db.get(MemberORM, (role, subject))
        if member_orm is None:
            subject_orm = db.get(SubjectORM, subject)
            if subject_orm is None:
                raise PyPermissionError(f"Subject '{subject}' does not exist!")
            role_orm = db.get(RoleORM, role)
            if role_orm is None:
                raise PyPermissionError(f"Role '{role}' does not exist!")
            raise PyPermissionError(
                f"Role '{role}' is not assigned to Subject '{subject}'!"
            )
        db.delete(member_orm)
        db.flush()

    @classmethod
    def roles(
        cls, *, subject: str, include_ascendant_roles: bool = False, db: Session
    ) -> tuple[str, ...]:
        """
        Get all Roles assigned to a Subject.

        Parameters
        ----------
        subject : str
            The target SubjectID.
        include_ascendant_roles: bool
            Include all ascendant Roles.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[str, ...]
            A tuple containing all assigned RoleIDs.

        Raises
        ------
        PyPermissionError
            If the target Subject does not exist.
        """
        if include_ascendant_roles:
            root_cte = (
                select(MemberORM.role_id)
                .where(MemberORM.subject_id == subject)
                .cte(recursive=True)
            )
            traversing_cte = root_cte.alias()
            relations_cte = root_cte.union_all(
                select(HierarchyORM.parent_role_id).join(
                    traversing_cte,
                    HierarchyORM.child_role_id == traversing_cte.c.role_id,
                )
            )
            selection = select(relations_cte)
            roles = db.scalars(selection).unique().all()
        else:
            roles = db.scalars(
                select(MemberORM.role_id).where(MemberORM.subject_id == subject)
            ).all()

        if len(roles) == 0 and db.get(SubjectORM, subject) is None:
            raise PyPermissionError(f"Subject '{subject}' does not exist!")
        return tuple(roles)

    @classmethod
    def check_permission(
        cls,
        *,
        subject: str,
        permission: Permission,
        db: Session,
    ) -> bool:
        """
        Check if a Subject has access to a specific Permission via its Role hierarchy.

        Parameters
        ----------
        subject : str
            The target SubjectID.
        permission : Permission
            The Permission to check for.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        bool
            True if the Permission is granted.

        Raises
        ------
        PyPermissionError
            If the target Subject does not exist. TODO
        """
        # TODO raise IntegrityError if subject is unknown and if possible via ORM

        root_cte = (
            select(MemberORM.role_id)
            .where(MemberORM.subject_id == subject)
            .cte(recursive=True)
        )

        traversing_cte = root_cte.alias()
        relations_cte = root_cte.union_all(
            select(HierarchyORM.parent_role_id).join(
                traversing_cte, HierarchyORM.child_role_id == traversing_cte.c.role_id
            )
        )

        policy_orms = db.scalars(
            select(PolicyORM)
            .join(relations_cte, PolicyORM.role_id == relations_cte.c.role_id)
            .where(
                PolicyORM.resource_type == permission.resource_type,
                PolicyORM.resource_id.in_((permission.resource_id, "*")),
                PolicyORM.action == permission.action,
            )
        ).all()
        if len(policy_orms) > 0:
            return True
        subject_orm = db.get(SubjectORM, subject)
        if subject_orm is None:
            raise PyPermissionError(f"Subject '{subject}' does not exist!")
        return False

    @classmethod
    def assert_permission(
        cls,
        *,
        subject: str,
        permission: Permission,
        db: Session,
    ) -> None:
        """
        Asserts that a Subject has access to a specific Permission via its Role hierarchy.

        Parameters
        ----------
        subject : str
            The target SubjectID.
        permission : Permission
            The Permission to check for.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionNotGrantedError
            If the Permission is not granted.
        PyPermissionError
            If the target Subject does not exist.
        """
        if not cls.check_permission(subject=subject, permission=permission, db=db):
            raise PermissionNotGrantedError(
                f"Permission '{permission}' is not granted for Subject '{subject}'!"
            )

    @classmethod
    def permissions(cls, *, subject: str, db: Session) -> tuple[Permission, ...]:
        """
        Get all Permissions a Subject has access to via its Role hierarchy.

        Parameters
        ----------
        subject : str
            The target SubjectID.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[Permission, ...]
            A tuple containing all granted Permissions.

        Raises
        ------
        PyPermissionError
            If the target Subject does not exist.
        """
        policy_orms = _get_policy_orms_for_subject(subject=subject, db=db)

        return tuple(
            Permission(
                resource_type=policy_orm.resource_type,
                resource_id=policy_orm.resource_id,
                action=policy_orm.action,
            )
            for policy_orm in policy_orms
        )

    @classmethod
    def policies(cls, *, subject: str, db: Session) -> tuple[Policy, ...]:
        """
        Get all Policies associated to a Subject via its Role hierarchy.

        Parameters
        ----------
        subject : str
            The target SubjectID.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[Policies, ...]
            A tuple containing all granted Policies.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
        """
        policy_orms = _get_policy_orms_for_subject(subject=subject, db=db)

        return tuple(
            Policy(
                role=policy_orm.role_id,
                permission=Permission(
                    resource_type=policy_orm.resource_type,
                    resource_id=policy_orm.resource_id,
                    action=policy_orm.action,
                ),
            )
            for policy_orm in policy_orms
        )

    @classmethod
    def actions_on_resource(
        cls,
        *,
        subject: str,
        resource_type: str,
        resource_id: str,
        inherited: bool = True,
        db: Session,
    ) -> tuple[str, ...]:
        if inherited:
            root_cte = (
                select(MemberORM.role_id)
                .where(MemberORM.subject_id == subject)
                .cte(recursive=True)
            )
            traversing_cte = root_cte.alias()
            relations_cte = root_cte.union_all(
                select(HierarchyORM.parent_role_id).join(
                    traversing_cte,
                    HierarchyORM.child_role_id == traversing_cte.c.role_id,
                )
            )
            actions = (
                db.scalars(
                    select(PolicyORM.action, PolicyORM.role_id)
                    .join(relations_cte, PolicyORM.role_id == relations_cte.c.role_id)
                    .where(
                        PolicyORM.resource_type == resource_type,
                        PolicyORM.resource_id.in_((resource_id, "*")),
                    )
                )
                .unique()
                .all()
            )
            tuple(actions)
        else:
            selection = (
                select(PolicyORM.action, PolicyORM.role_id)
                .join(MemberORM, MemberORM.role_id == PolicyORM.role_id)
                .where(
                    MemberORM.subject_id == subject,
                    PolicyORM.resource_type == resource_type,
                    PolicyORM.resource_id.in_((resource_id, "*")),
                )
            )
            actions = db.scalars(selection).unique().all()
        if len(actions) == 0 and db.get(SubjectORM, subject) is None:
            raise PyPermissionError(f"Subject '{subject}' does not exist!")
        return tuple(actions)


################################################################################
#### Util
################################################################################


def _get_policy_orms_for_subject(*, subject: str, db: Session) -> Sequence[PolicyORM]:
    subject_orm = db.get(SubjectORM, subject)
    if not subject_orm:
        raise PyPermissionError(f"Subject '{subject}' does not exist!")
    root_cte = (
        select(MemberORM.role_id)
        .where(MemberORM.subject_id == subject)
        .cte(recursive=True)
    )

    traversing_cte = root_cte.alias()
    relations_cte = root_cte.union_all(
        select(HierarchyORM.parent_role_id).join(
            traversing_cte, HierarchyORM.child_role_id == traversing_cte.c.role_id
        )
    )

    policy_orms = (
        db.scalars(
            select(PolicyORM).join(
                relations_cte, PolicyORM.role_id == relations_cte.c.role_id
            )
        )
        .unique()
        .all()
    )

    return policy_orms



from typing import Final

from sqlalchemy.engine.base import Engine

from pypermission import RBAC
from pypermission.models import Policy, Permission, create_rbac_database_table
from pypermission.example.models import Context
from pypermission.example.service.user import UserService
from pypermission.example.service.group import GroupService


class ExampleApp:
    _user: Final = UserService
    _group: Final = GroupService

    def __init__(self, *, engine: Engine) -> None:
        create_rbac_database_table(engine=engine)

    @property
    def user(self) -> type[UserService]:
        return self._user

    @property
    def group(self) -> type[GroupService]:
        return self._group

    def populate_example(self, *, ctx: Context) -> None:
        RBAC.role.create(role="admin", db=ctx.db)
        RBAC.role.create(role="moderator", db=ctx.db)
        RBAC.role.create(role="user", db=ctx.db)
        RBAC.role.create(role="guest", db=ctx.db)

        RBAC.role.add_hierarchy(
            parent_role="guest",
            child_role="user",
            db=ctx.db,
        )
        RBAC.role.add_hierarchy(
            parent_role="user",
            child_role="moderator",
            db=ctx.db,
        )
        RBAC.role.add_hierarchy(
            parent_role="moderator",
            child_role="admin",
            db=ctx.db,
        )

        self._create_guest_role_policies(ctx=ctx)
        self._create_user_role_policies(ctx=ctx)
        self._create_moderator_role_policies(ctx=ctx)
        self._create_admin_role_policies(ctx=ctx)

        ctx.db.flush()

        self.user.create(
            username="Alex",
            email="alex@digon.io",
            role="admin",
            ctx=ctx,
            rbac=False,
        )
        self.user.create(
            username="Max",
            email="max@digon.io",
            role="moderator",
            ctx=ctx,
            rbac=False,
        )
        self.user.create(
            username="Ulrich",
            email="urlich@digon.io",
            role="user",
            ctx=ctx,
            rbac=False,
        )
        self.user.create(
            username="Uwe",
            email="Uwe@digon.io",
            role="user",
            ctx=ctx,
            rbac=False,
        )
        self.user.create(
            username="Georg",
            email="Georg@digon.io",
            role="guest",
            ctx=ctx,
            rbac=False,
        )

        self.group.create(
            groupname="BEF",
            description="Bergisches Entwicklerforum",
            owner="Alex",
            ctx=ctx,
            rbac=False,
        )
        self.group.create(
            groupname="PUGW",
            description="Python User Group Wuppertal",
            owner="Uwe",
            ctx=ctx,
            rbac=False,
        )
        ctx.db.flush()

    def _create_guest_role_policies(self, ctx: Context) -> None:
        RBAC.role.grant_permission(
            role="guest",
            permission=Permission(
                resource_type="group",
                resource_id="*",
                action="access",
            ),
            db=ctx.db,
        )

    def _create_user_role_policies(self, ctx: Context) -> None:
        RBAC.role.grant_permission(
            role="user",
            permission=Permission(
                resource_type="base",
                resource_id="",
                action="group:create",
            ),
            db=ctx.db,
        )

    def _create_moderator_role_policies(self, ctx: Context) -> None:
        RBAC.role.grant_permission(
            role="moderator",
            permission=Permission(
                resource_type="base",
                resource_id="",
                action="user:create",
            ),
            db=ctx.db,
        )
        RBAC.role.grant_permission(
            role="moderator",
            permission=Permission(
                resource_type="user",
                resource_id="*",
                action="access",
            ),
            db=ctx.db,
        )
        RBAC.role.grant_permission(
            role="moderator",
            permission=Permission(
                resource_type="user",
                resource_id="*",
                action="edit",
            ),
            db=ctx.db,
        )
        RBAC.role.grant_permission(
            role="moderator",
            permission=Permission(
                resource_type="user",
                resource_id="*",
                action="deactivate",
            ),
            db=ctx.db,
        )
        RBAC.role.grant_permission(
            role="moderator",
            permission=Permission(
                resource_type="group",
                resource_id="*",
                action="deactivate",
            ),
            db=ctx.db,
        )

    def _create_admin_role_policies(self, ctx: Context) -> None:
        RBAC.role.grant_permission(
            role="admin",
            permission=Permission(
                resource_type="base",
                resource_id="",
                action="rbac",
            ),
            db=ctx.db,
        )
        RBAC.role.grant_permission(
            role="admin",
            permission=Permission(
                resource_type="user",
                resource_id="*",
                action="delete",
            ),
            db=ctx.db,
        )
        RBAC.role.grant_permission(
            role="admin",
            permission=Permission(
                resource_type="group",
                resource_id="*",
                action="delete",
            ),
            db=ctx.db,
        )



from typing import Sequence

from sqlalchemy.sql import select

from pypermission import RBAC, Permission
from pypermission.example.models import UserORM, Context, ExampleError, State

################################################################################
#### UserService
################################################################################


class UserService:
    @classmethod
    def create(
        cls,
        *,
        username: str,
        email: str,
        role: str = "guest",
        ctx: Context,
        rbac: bool = True,
    ) -> UserORM:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="base", resource_id="", action="user:create"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        user_orm = UserORM(username=username, email=email, role=role)
        ctx.db.add(user_orm)
        ctx.db.flush()

        cls._create_role_and_policies(username=username, role=role, ctx=ctx)

        return user_orm

    @classmethod
    def list(
        cls,
        *,
        ctx: Context,
        rbac: bool = True,
    ) -> Sequence[UserORM]:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="user", resource_id="*", action="access"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        return ctx.db.scalars(select(UserORM)).all()

    @classmethod
    def get(
        cls,
        *,
        username: str,
        ctx: Context,
        rbac: bool = True,
    ) -> UserORM:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="user", resource_id=username, action="access"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        user_orm = ctx.db.get(UserORM, username)
        if user_orm is None:
            raise ExampleError(f"Unknown user '{username}'!")
        return user_orm

    @classmethod
    def set_email(
        cls,
        *,
        username: str,
        email: str,
        ctx: Context,
        rbac: bool = True,
    ) -> UserORM:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="user", resource_id=username, action="edit"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        user_orm = ctx.db.get(UserORM, username)
        if user_orm is None:
            raise ExampleError(f"Unknown user '{username}'!")
        user_orm.email = email
        ctx.db.flush()
        return user_orm

    @classmethod
    def set_state(
        cls,
        *,
        username: str,
        state: State,
        ctx: Context,
        rbac: bool = True,
    ) -> UserORM:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="user", resource_id=username, action="deactivate"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        user_orm = ctx.db.get(UserORM, username)
        if user_orm is None:
            raise ExampleError(f"Unknown user '{username}'!")
        user_orm.state = state
        ctx.db.flush()
        return user_orm

    @classmethod
    def delete(
        cls,
        *,
        username: str,
        ctx: Context,
        rbac: bool = True,
    ) -> UserORM:

        match rbac, ctx.username:
            case True, str():
                # NOTE Only needed, because no real authentication is present.
                if ctx.db.get(UserORM, ctx.username) is None:
                    raise ExampleError(f"Unknown user '{ctx.username}' in context!")

                if not RBAC.subject.check_permission(
                    subject=ctx.username,
                    permission=Permission(
                        resource_type="user", resource_id=username, action="delete"
                    ),
                    db=ctx.db,
                ):
                    raise ExampleError("Permission not granted!")
            case False, None:
                ...
            case False, str():
                ...
            case True, None:
                raise ExampleError("No user in context!")

        user_orm = ctx.db.get(UserORM, username)
        if user_orm is None:
            raise ExampleError(f"Unknown user '{username}'!")

        USER_ROLE = f"user[{username}]"
        RBAC.role.delete(role=USER_ROLE, db=ctx.db)
        RBAC.subject.delete(subject=username, db=ctx.db)

        for group_orm in user_orm.group_orms:
            GROUP_ROLE__OWNER = f"group[{group_orm.groupname}]:owner"
            RBAC.role.delete(role=GROUP_ROLE__OWNER, db=ctx.db)

        ctx.db.delete(user_orm)
        ctx.db.flush()

        return user_orm

    ################################################################################
    #### Util
    ################################################################################

    @classmethod
    def _create_role_and_policies(cls, username: str, role: str, ctx: Context) -> None:
        RBAC.subject.create(subject=username, db=ctx.db)
        RBAC.subject.assign_role(role=role, subject=username, db=ctx.db)

        USER_ROLE = f"user[{username}]"
        RBAC.role.create(role=USER_ROLE, db=ctx.db)
        RBAC.subject.assign_role(subject=username, role=USER_ROLE, db=ctx.db)

        RBAC.role.grant_permission(
            role=USER_ROLE,
            permission=Permission(
                resource_type="user", resource_id=username, action="access"
            ),
            db=ctx.db,
        )
        RBAC.role.grant_permission(
            role=USER_ROLE,
            permission=Permission(
                resource_type="user", resource_id=username, action="edit"
            ),
            db=ctx.db,
        )
        RBAC.role.grant_permission(
            role=USER_ROLE,
            permission=Permission(
                resource_type="user", resource_id=username, action="deactivate"
            ),
            db=ctx.db,
        )



import networkx as nx
import plotly.graph_objects as go

from pypermission.exc import PyPermissionError


def plot_factory(
    *, dag: nx.DiGraph, auto_open: bool = False, file_path: str = "dag.html"
) -> None:
    """
    Generate an interactive HTML visualization of an RBAC DAG and save it to `file_path`.

    Parameters
    ----------
    dag : nx.DiGraph
        The RBAC system DAG.
    auto_open : bool
        Automatically opens the plot in the browser.
    file_path: str
        Path and filename for the generated HTML as string.

    Raises
    ------
    PyPermissionError
        Raised when the input DAG is empty, indicating that there is no data to plot.
    """

    if not len(dag):
        raise PyPermissionError("RBAC system is empty. Nothing to plot!")

    fig = _build_plotly_figure(dag=dag)
    fig.write_html(file_path, auto_open=auto_open)


################################################################################
#### Util
################################################################################

COLOR_MAP = {
    "subject_node": "darkturquoise",
    "member_edge": "darkturquoise",
    "role_node": "coral",
    "hierarchy_edge": "coral",
    "permission_node": "forestgreen",
    "policy_edge": "forestgreen",
}

NodePositions = dict[str, tuple[float, float]]


def _build_plotly_figure(*, dag: nx.DiGraph) -> go.Figure:
    node_positions = _calc_node_positions(dag=dag)
    nodes = _build_nodes(dag=dag, node_positions=node_positions)

    edge_colors = tuple(COLOR_MAP[dag.edges[n]["type"]] for n in dag.edges())
    edges = _build_edges(
        dag=dag, node_positions=node_positions, edge_colors=edge_colors
    )

    fig = go.Figure(data=[edges[0], edges[1], edges[2], nodes[0], nodes[1], nodes[2]])
    fig.update_layout(
        showlegend=True,
        margin=dict(l=20, r=20, t=20, b=20),
        xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
        yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
    )

    return fig


def _build_nodes(
    *, dag: nx.DiGraph, node_positions: NodePositions
) -> tuple[go.Scatter, go.Scatter, go.Scatter]:
    node_subjects = tuple(
        node for node, data in dag.nodes.data() if data["type"] == "subject_node"
    )
    node_roles = tuple(
        node for node, data in dag.nodes.data() if data["type"] == "role_node"
    )
    node_permissions = tuple(
        node for node, data in dag.nodes.data() if data["type"] == "permission_node"
    )

    return (
        go.Scatter(
            x=tuple(node_positions[node][0] for node in node_subjects),
            y=tuple(node_positions[node][1] for node in node_subjects),
            mode="markers+text",
            text=tuple(str(n) for n in node_subjects),
            textposition="top center",
            textfont=dict(size=20, color="black"),
            marker=dict(size=20, color=COLOR_MAP["subject_node"]),
            name="Subjects",
        ),
        go.Scatter(
            x=tuple(node_positions[node][0] for node in node_roles),
            y=tuple(node_positions[node][1] for node in node_roles),
            mode="markers+text",
            text=[str(n) for n in node_roles],
            textposition="top center",
            textfont=dict(size=20, color="black"),
            marker=dict(size=20, color=COLOR_MAP["role_node"]),
            name="Roles",
        ),
        go.Scatter(
            x=tuple(node_positions[node][0] for node in node_permissions),
            y=tuple(node_positions[node][1] for node in node_permissions),
            mode="markers+text",
            text=tuple(str(node) for node in node_permissions),
            textposition="top center",
            textfont=dict(size=20, color="black"),
            marker=dict(size=20, color=COLOR_MAP["permission_node"]),
            name="Permissions",
        ),
    )


def _build_edges(
    *, dag: nx.DiGraph, node_positions: NodePositions, edge_colors: tuple[str, ...]
) -> tuple[go.Scatter, go.Scatter, go.Scatter]:
    edge_x_member, edge_y_member = [], []
    edge_x_hierarchy, edge_y_hierarchy = [], []
    edge_x_policy, edge_y_policy = [], []

    for n_0, n_1, data in dag.edges.data():
        x_0, y_0 = node_positions[n_0]
        x_1, y_1 = node_positions[n_1]

        match data["type"]:
            case "member_edge":
                edge_x_member.extend([x_0, x_1, None])
                edge_y_member.extend([y_0, y_1, None])
            case "hierarchy_edge":
                edge_x_hierarchy.extend([x_0, x_1, None])
                edge_y_hierarchy.extend([y_0, y_1, None])
            case "policy_edge":
                edge_x_policy.extend([x_0, x_1, None])
                edge_y_policy.extend([y_0, y_1, None])

    return (
        go.Scatter(
            x=edge_x_member,
            y=edge_y_member,
            line=dict(width=2, color=COLOR_MAP["member_edge"]),
            hoverinfo="none",
            mode="lines",
            name="Role assignment",
        ),
        go.Scatter(
            x=edge_x_hierarchy,
            y=edge_y_hierarchy,
            line=dict(width=2, color=COLOR_MAP["hierarchy_edge"]),
            hoverinfo="none",
            mode="lines",
            name="Role hierarchy",
        ),
        go.Scatter(
            x=edge_x_policy,
            y=edge_y_policy,
            line=dict(width=2, color=COLOR_MAP["policy_edge"]),
            hoverinfo="none",
            mode="lines",
            name="Permission assignment",
        ),
    )


def _calc_node_positions(*, dag: nx.DiGraph) -> dict[str, tuple[float, float]]:

    role_nodes = [n for n, d in dag.nodes(data=True) if d.get("type") == "role_node"]
    hierarchy_edges = [
        (u, v) for u, v, d in dag.edges(data=True) if d.get("type") == "hierarchy_edge"
    ]

    role_sdag = nx.DiGraph()
    role_sdag.add_nodes_from(role_nodes)
    role_sdag.add_edges_from(hierarchy_edges)

    unsorted_role_sdags = [
        dag.subgraph(c).copy() for c in nx.weakly_connected_components(role_sdag)
    ]
    role_sdags = sorted(
        unsorted_role_sdags, key=lambda g: g.number_of_nodes(), reverse=False
    )

    role_layers: dict[str, int] = {}
    for sdag in role_sdags:
        for node in nx.topological_sort(sdag):
            predecessors = tuple(dag.predecessors(node))
            role_layers[node] = (
                1
                + max(
                    tuple(
                        role_layers[p]
                        for p in predecessors
                        if dag.nodes[p]["type"] == "role_node"
                    )
                    + (0,)
                )
                if predecessors
                else 1
            )

    permission_layer = 0
    subject_layer = max(role_layers.values()) + 1

    layer_x_nodes: dict[int, list[str]] = {
        layer: [] for layer in range(permission_layer, subject_layer + 1)
    }
    for node in nx.topological_sort(dag):
        match dag.nodes[node]["type"]:
            case "subject_node":
                layer_x_nodes[subject_layer].append(node)
            case "permission_node":
                layer_x_nodes[permission_layer].append(node)

    for sdag in role_sdags:
        for node in nx.topological_sort(sdag):
            layer_x_nodes[role_layers[node]].append(node)

    node_positions = {}
    for layer, nodes_in_layer in layer_x_nodes.items():
        n_nodes = len(nodes_in_layer)
        ys: tuple[float, ...]
        if n_nodes == 1:
            ys = (1,)
        else:
            ys = tuple(1 * x / (n_nodes - 1) for x in range(n_nodes))
        x = -layer
        for y, node in zip(ys, nodes_in_layer):
            node_positions[node] = (float(x), y)
    return node_positions



from enum import StrEnum

from sqlalchemy.orm import Mapped, mapped_column, relationship, Session
from sqlalchemy.sql.sqltypes import String
from sqlalchemy.sql.sqltypes import Enum as SqlEnum
from sqlalchemy.sql.schema import ForeignKey


from pypermission.models import BaseORM

################################################################################
#### Types
################################################################################


class ExampleError(Exception): ...


class Context:
    username: str | None
    db: Session

    def __init__(self, *, user: str | None = None, db: Session):
        self.username = user
        self.db = db


class State(StrEnum):
    ACTIVE = "active"
    INACTIVE = "inactive"


################################################################################
#### UserORM
################################################################################


class UserORM(BaseORM):
    __tablename__ = "app_user_table"
    username: Mapped[str] = mapped_column(String, primary_key=True)
    email: Mapped[str] = mapped_column(String)
    role: Mapped[str] = mapped_column(String)
    state: Mapped[State] = mapped_column(
        SqlEnum(State, name="UserORM.State"), default=State.ACTIVE
    )

    group_orms: Mapped[list["GroupORM"]] = relationship(
        "GroupORM", back_populates="owner_orm"
    )


################################################################################
#### GroupORM
################################################################################


class GroupORM(BaseORM):
    __tablename__ = "app_group_table"
    groupname: Mapped[str] = mapped_column(String, primary_key=True)
    description: Mapped[str] = mapped_column(String)
    owner: Mapped[str] = mapped_column(
        String,
        ForeignKey("app_user_table.username", ondelete="CASCADE"),
    )
    state: Mapped[State] = mapped_column(
        SqlEnum(State, name="GroupORM.State"), default=State.ACTIVE
    )

    owner_orm: Mapped["UserORM"] = relationship("UserORM", back_populates="group_orms")



from sqlalchemy.orm import Mapped, mapped_column, DeclarativeBase
from sqlalchemy.sql.sqltypes import String
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy.engine.base import Engine
from typing import Never
from sqlite3 import Connection
from sqlalchemy.pool.base import (
    _ConnectionRecord,  # pyright: ignore[reportPrivateUsage]
)


class BaseORM(DeclarativeBase): ...


################################################################################
#### Types
################################################################################


class Permission:
    """
    Represents a Resource paired with an Action.

    Attributes
    ----------
    resource_type : str
        The ResourceType (e.g., "document", "user").
    resource_id : str
        The ResourceID. The star '*' acts as a wildcard matching all ResourceIDs of the same ResourceType. The empty string can be used for Actions on Resources that do not have an ResourceID.
    action : str
        The Action allowed on the Resource (e.g., "read", "write", "delete").
    """

    resource_type: str
    resource_id: str
    action: str

    def __init__(self, *, resource_type: str, resource_id: str, action: str) -> None:
        """
        Initialize the Permission.

        Parameters
        ----------
        resource_type : str
            The type of the resource (e.g., "document", "user").
        resource_id : str
            The ID of the resource instance. The start '*' acts as a wildcard matching all IDs of the resource. The empty string can be used for actions on resources that do not have an ID.
        action : str
            The action allowed on the resource (e.g., "read", "write", "delete").
        """
        self.resource_type = resource_type
        self.resource_id = resource_id
        self.action = action

    def __str__(self) -> str:
        if not self.resource_id:
            return f"{self.resource_type}:{self.action}"
        return f"{self.resource_type}[{self.resource_id}]:{self.action}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Permission):
            return False

        return (
            self.resource_type == other.resource_type
            and self.resource_id == other.resource_id
            and self.action == other.action
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


class Policy:
    """
    Represents a Role paired with a Permission.

    Attributes
    ----------
    role : str
        The target RoleID.
    permission : Permission
        The target Permission.
    """

    role: str
    permission: Permission

    def __init__(self, *, role: str, permission: Permission) -> None:
        """
        Initialize the Policy.

        Parameters
        ----------
        role : str
            The target RoleID.
        permission : Permission
            The target Permission.
        """
        self.role = role
        self.permission = permission

    def __str__(self) -> str:
        return f"{self.role}:{self.permission}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Policy):
            return False

        return self.role == other.role and self.permission == other.permission

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


class FrozenClass(type):
    def __setattr__(cls, key: str, value: Never) -> None:
        if key in cls.__dict__:
            raise AttributeError(f"FrozenClass attributes cannot be overwrite!")
        super().__setattr__(key, value)


################################################################################
#### RoleORM
################################################################################


class RoleORM(BaseORM):
    __tablename__ = "pp_role_table"
    id: Mapped[str] = mapped_column(String, primary_key=True)


################################################################################
#### HierarchyORM
################################################################################


class HierarchyORM(BaseORM):
    __tablename__ = "pp_hierarchy_table"
    parent_role_id: Mapped[str] = mapped_column(
        String, ForeignKey("pp_role_table.id", ondelete="CASCADE"), primary_key=True
    )
    child_role_id: Mapped[str] = mapped_column(
        String, ForeignKey("pp_role_table.id", ondelete="CASCADE"), primary_key=True
    )


################################################################################
#### SubjectORM
################################################################################


class SubjectORM(BaseORM):
    __tablename__ = "pp_subject_table"
    id: Mapped[str] = mapped_column(String, primary_key=True)


################################################################################
#### MemberORM
################################################################################


class MemberORM(BaseORM):
    __tablename__ = "pp_member_table"
    role_id: Mapped[str] = mapped_column(
        String, ForeignKey("pp_role_table.id", ondelete="CASCADE"), primary_key=True
    )
    subject_id: Mapped[str] = mapped_column(
        String, ForeignKey("pp_subject_table.id", ondelete="CASCADE"), primary_key=True
    )


################################################################################
#### PolicyORM
################################################################################


class PolicyORM(BaseORM):
    __tablename__ = "pp_policy_table"
    role_id: Mapped[str] = mapped_column(
        String, ForeignKey("pp_role_table.id", ondelete="CASCADE"), primary_key=True
    )
    resource_type: Mapped[str] = mapped_column(String, primary_key=True)
    resource_id: Mapped[str] = mapped_column(String, primary_key=True)
    action: Mapped[str] = mapped_column(String, primary_key=True)


################################################################################
#### Util
################################################################################


def create_rbac_database_table(*, engine: Engine) -> None:
    """
    Create all required database tables via. SQLAlchemy for PyPermission.

    Parameters
    ----------
    engine : Engine
        The SQLAlchemy engine.

    """
    BaseORM.metadata.create_all(bind=engine)


# https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#foreign-key-support
def set_sqlite_pragma(
    dbapi_connection: Connection, _connection_record: _ConnectionRecord
) -> None:
    # the sqlite3 driver will not set PRAGMA foreign_keys
    # if autocommit=False; set to True temporarily
    ac = dbapi_connection.autocommit
    dbapi_connection.autocommit = True

    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()

    # restore previous autocommit setting
    dbapi_connection.autocommit = ac



from typing import Never
from sqlalchemy.exc import IntegrityError
from sqlite3 import IntegrityError as Sqlite3IntegrityError
from pypermission.models import Permission

################################################################################
#### Generic Errors
################################################################################


class PyPermissionError(Exception):
    """
    PyPermissionError is the standard error of PyPermission.

    Attributes
    ----------
    message : str
        A detailed description of the occurred error.
    """

    message: str

    def __init__(self, message: str = ""):
        self.message = message


class PermissionNotGrantedError(PyPermissionError):
    """
    PermissionNotGrantedError will be thrown if an `assert_permission()` fails!

    Attributes
    ----------
    message : str
        A constant error description.
    """

    message = "RBAC: Permission not granted!"


################################################################################
#### Error Utilities
################################################################################


def process_subject_role_integrity_error(
    *, err: IntegrityError, subject: str | None = None, role: str | None = None
) -> Never:
    match err:
        case IntegrityError(orig=Sqlite3IntegrityError()):
            if subject is not None and role is not None:
                raise PyPermissionError(
                    f"Subject '{subject}' or Role '{role}' does not exist!"
                )
            if subject is not None:
                raise PyPermissionError(f"Subject '{subject}' does not exist!")
            if role is not None:
                raise PyPermissionError(f"Role '{role}' does not exist!")
        case IntegrityError(orig=orig) if (
            str(orig.__class__) == "<class 'psycopg.errors.ForeignKeyViolation'>"
        ):
            if f"Key (role_id)=({role}) is not present in table" in str(orig):
                raise PyPermissionError(f"Role '{role}' does not exist!")
            if f"Key (subject_id)=({subject}) is not present in table" in str(orig):
                raise PyPermissionError(f"Subject '{subject}' does not exist!")
        case _:
            ...
    raise PyPermissionError("Unexpected IntegrityError")


def process_policy_integrity_error(
    *,
    err: IntegrityError,
    role: str,
    permission: Permission,
) -> Never:
    match err:
        case IntegrityError(
            orig=Sqlite3IntegrityError(sqlite_errorname="SQLITE_CONSTRAINT_PRIMARYKEY")
        ):
            raise PyPermissionError(
                f"Permission '{str(permission)}' does already exist!"
            )
        case IntegrityError(
            orig=Sqlite3IntegrityError(sqlite_errorname="SQLITE_CONSTRAINT_FOREIGNKEY")
        ):
            raise PyPermissionError(f"Role '{role}' does not exist!")
        case IntegrityError(orig=orig) if (
            str(orig.__class__) == "<class 'psycopg.errors.UniqueViolation'>"
        ):
            raise PyPermissionError(
                f"Permission '{str(permission)}' does already exist!"
            )
        case IntegrityError(orig=orig) if (
            str(orig.__class__) == "<class 'psycopg.errors.ForeignKeyViolation'>"
        ):
            raise PyPermissionError(f"Role '{role}' does not exist!")
        case _:
            ...
    raise PyPermissionError("Unexpected IntegrityError")


################################################################################
#### Testing only:
################################################################################


class ERR_MSG:
    role_exists = "Role '{role}' already exists!"
    subject_exists = "Subject '{subject}' already exists!"
    hierarchy_exists = "Hierarchy '{parent_role}' -> '{child_role}' exists!"
    permission_exists = "Permission '{permission_str}' does already exist!"

    non_existent_subject_role = "Subject '{subject}' or Role '{role}' does not exist!"
    non_existent_subject = "Subject '{subject}' does not exist!"
    non_existent_role = "Role '{role}' does not exist!"
    non_existent_hierarchy = (
        "Hierarchy '{parent_role}' -> '{child_role}' does not exist!"
    )
    non_existent_parent_child_roles = (
        "Roles '{parent_role}' and '{child_role}' do not exist!"
    )
    non_existent_role_assignment = (
        "Role '{role}' is not assigned to Subject '{subject}'!"
    )
    non_existent_permission = "Permission '{permission_str}' does not exist!"

    permission_not_granted_for_role = (
        "Permission '{permission_str}' is not granted for Role '{role}'!"
    )
    permission_not_granted_for_subject = (
        "Permission '{permission_str}' is not granted for Subject '{subject}'!"
    )

    conflicting_role_ids = "RoleIDs must not be equal: '{role}'!"
    cycle_detected = "Desired hierarchy would create a cycle!"
    unexpected_integrity = "Unexpected IntegrityError!"



import networkx as nx
import plotly.graph_objects as go

from sqlalchemy.orm import Session
from sqlalchemy.sql import select

from pypermission.models import RoleORM, HierarchyORM, MemberORM, PolicyORM
from pypermission.exc import PyPermissionError

################################################################################
#### Role dag tools
################################################################################


def role_dag(
    *,
    root_roles: set[str] | None = None,
    include_subjects: bool = True,
    include_permissions: bool = True,
    db: Session,
) -> nx.DiGraph:
    """
    Generate a directed acyclic graph (DAG) that reflects the role hierarchy
    from the viewpoint of the `root_roles`.
    By definition, the resulting DAG does **not** contain any roles that are
    descendants of those root roles.

    Parameters
    ----------
    root_roles : set[str] | None
        Root Roles of the generated DAG. If its None, all existing Roles will be included.
    include_subjects : bool
        Include assigned Subjects in the DAG.
    include_permissions : bool
        Include granted Permissions in the DAG.

    Returns
    -------
    nx.DiGraph
        The generated DAG.
    """
    roles, hierarchies = _get_roles_and_hierarchies(root_roles=root_roles, db=db)

    dag = nx.DiGraph()
    dag.add_nodes_from(roles, type="role_node")
    dag.add_edges_from(hierarchies, type="hierarchy_edge")

    if include_subjects:
        subjects, members = _get_subjects_and_members(roles=roles, db=db)
        dag.add_nodes_from(subjects, type="subject_node")
        dag.add_edges_from(members, type="member_edge")

    if include_permissions:
        subjects, members = _get_permissions_and_polices(roles=roles, db=db)
        dag.add_nodes_from(subjects, type="permission_node")
        dag.add_edges_from(members, type="policy_edge")

    return dag


################################################################################
#### Util
################################################################################


def _get_roles_and_hierarchies(
    *, root_roles: set[str] | None, db: Session
) -> tuple[set[str], set[tuple[str, str]]]:
    if root_roles is None:
        role_orms = db.scalars(select(RoleORM)).all()
        roles = set(role_orm.id for role_orm in role_orms)

        hierarchy_orms = db.scalars(select(HierarchyORM)).all()
        hierarchies = set(
            (hierarchy_orm.parent_role_id, hierarchy_orm.child_role_id)
            for hierarchy_orm in hierarchy_orms
        )
        return roles, hierarchies

    role_ormes = db.scalars(select(RoleORM).where(RoleORM.id.in_(root_roles))).all()
    roles = set(role_orme.id for role_orme in role_ormes)

    if unknown_roles := root_roles ^ roles:
        if len(unknown_roles) == 1:
            raise PyPermissionError(f"Requested role does not exist: {unknown_roles}!")
        else:
            raise PyPermissionError(f"Requested roles do not exist: {unknown_roles}!")

    root_cte = (
        select(HierarchyORM)
        .where(HierarchyORM.child_role_id.in_(root_roles))
        .cte(name="root_cte", recursive=True)
    )

    traversing_cte = root_cte.alias()
    relations_cte = root_cte.union_all(
        select(HierarchyORM).where(
            HierarchyORM.child_role_id == traversing_cte.c.parent_role_id
        )
    )

    ancestor_relations = db.execute(
        select(relations_cte.c.parent_role_id, relations_cte.c.child_role_id)
    ).all()

    roles = root_roles | {role for pair in ancestor_relations for role in pair}
    hierarchies = set(ancestor_relations)  # type: ignore

    return roles, hierarchies


def _get_subjects_and_members(
    *, roles: set[str], db: Session
) -> tuple[set[str], set[tuple[str, str]]]:

    member_orms = db.scalars(
        select(MemberORM).where(MemberORM.role_id.in_(roles))
    ).all()
    members = set(
        (member_orm.role_id, member_orm.subject_id) for member_orm in member_orms
    )

    subjects = {member_orm.subject_id for member_orm in member_orms}

    return subjects, members


def _get_permissions_and_polices(
    *, roles: set[str], db: Session
) -> tuple[set[str], set[tuple[str, str]]]:

    policy_orms = db.scalars(
        select(PolicyORM).where(PolicyORM.role_id.in_(roles))
    ).all()
    policies = set(
        (
            policy_orm.role_id,
            _permission_to_str(
                policy_orm.resource_type, policy_orm.resource_id, policy_orm.action
            ),
        )
        for policy_orm in policy_orms
    )

    permissions = set(policy[1] for policy in policies)

    return permissions, policies


def _permission_to_str(resource_type: str, resource_id: str, action: str) -> str:
    if not resource_id:
        return f"{resource_type}:{action}"
    return f"{resource_type}[{resource_id}]:{action}"



from collections.abc import Sequence

from sqlalchemy.sql import select
from sqlalchemy.orm import Session
from sqlalchemy.exc import IntegrityError

from pypermission.models import (
    Permission,
    RoleORM,
    HierarchyORM,
    PolicyORM,
    FrozenClass,
    MemberORM,
    Policy,
)
from pypermission.exc import (
    PyPermissionError,
    PermissionNotGrantedError,
    process_policy_integrity_error,
)

################################################################################
#### RoleService
################################################################################


class RoleService(metaclass=FrozenClass):

    @classmethod
    def create(cls, *, role: str, db: Session) -> None:
        """
        Create a new Role.

        Parameters
        ----------
        role : str
            The RoleID of the Role to create.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If a Role with the given RoleID already exists.
        """
        try:
            role_orm = RoleORM(id=role)
            db.add(role_orm)
            db.flush()
        except IntegrityError:
            db.rollback()
            raise PyPermissionError(f"Role '{role}' already exists!")

    @classmethod
    def delete(cls, *, role: str, db: Session) -> None:
        """
        Delete an existing Role.

        Parameters
        ----------
        role : str
            The RoleID to delete.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If a Role with the given RoleID does not exist.
        """
        role_orm = db.get(RoleORM, role)
        if role_orm is None:
            raise PyPermissionError(f"Role '{role}' does not exist!")
        db.delete(role_orm)
        db.flush()

    @classmethod
    def list(cls, *, db: Session) -> tuple[str, ...]:
        """
        Get all Roles.

        Parameters
        ----------
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[str, ...]
            A tuple containing all RoleIDs.
        """
        role_orms = db.scalars(select(RoleORM)).all()
        return tuple(role_orm.id for role_orm in role_orms)

    @classmethod
    def add_hierarchy(cls, *, parent_role: str, child_role: str, db: Session) -> None:
        """
        Add a parent-child hierarchy between two Roles.

        Parameters
        ----------
        parent_role : str
            The parent RoleID.
        child_role : str
            The child RoleID.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If arguments `parent_role` and `child_role` are equal.
            If one or both Roles do not exist.
            If adding the hierarchy would create a cycle.
            If the hierarchy already exists.
        """
        if parent_role == child_role:
            raise PyPermissionError(f"RoleIDs must not be equal: '{parent_role}'!")

        roles = db.scalars(
            select(RoleORM.id).where(RoleORM.id.in_([parent_role, child_role]))
        ).all()
        if len(roles) == 1:
            missing_role = child_role if parent_role in roles else parent_role
            raise PyPermissionError(f"Role '{missing_role}' does not exist!")
        elif len(roles) == 0:
            raise PyPermissionError(
                f"Roles '{parent_role}' and '{child_role}' do not exist!"
            )

        root_cte = (
            select(HierarchyORM)
            .where(HierarchyORM.parent_role_id == child_role)
            .cte(recursive=True)
        )

        traversing_cte = root_cte.alias()
        relations_cte = root_cte.union_all(
            select(HierarchyORM).where(
                HierarchyORM.parent_role_id == traversing_cte.c.child_role_id
            )
        )

        critical_leaf_relations = db.execute(
            select(relations_cte).where(relations_cte.c.child_role_id == parent_role)
        ).all()

        if critical_leaf_relations:
            raise PyPermissionError("Desired hierarchy would create a cycle!")

        try:
            hierarchy_orm = HierarchyORM(
                parent_role_id=parent_role, child_role_id=child_role
            )
            db.add(hierarchy_orm)
            db.flush()
        except IntegrityError as err:
            db.rollback()
            raise PyPermissionError(
                f"Hierarchy '{parent_role}' -> '{child_role}' exists!"
            )

    @classmethod
    def remove_hierarchy(
        cls, *, parent_role: str, child_role: str, db: Session
    ) -> None:
        """
        Remove a parent-child hierarchy between two Roles.

        Parameters
        ----------
        parent_role : str
            The parent RoleID.
        child_role : str
            The child RoleID.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If arguments `parent_role` and `child_role` are equal.
            If one or both Roles do not exist.
            If the hierarchy does not exist.
        """
        if parent_role == child_role:
            raise PyPermissionError(f"RoleIDs must not be equal: '{parent_role}'!")

        hierarchy_orm = db.get(HierarchyORM, (parent_role, child_role))
        if hierarchy_orm is None:
            roles = db.scalars(
                select(RoleORM.id).where(RoleORM.id.in_([parent_role, child_role]))
            ).all()
            if len(roles) == 1:
                missing_role = child_role if parent_role in roles else parent_role
                raise PyPermissionError(f"Role '{missing_role}' does not exist!")
            elif len(roles) == 0:
                raise PyPermissionError(
                    f"Roles '{parent_role}' and '{child_role}' do not exist!"
                )
            else:
                raise PyPermissionError(
                    f"Hierarchy '{parent_role}' -> '{child_role}' does not exist!"
                )

        db.delete(hierarchy_orm)
        db.flush()

    @classmethod
    def parents(cls, *, role: str, db: Session) -> tuple[str, ...]:
        """
        Get all parent Roles.

        Parameters
        ----------
        role : str
            The target RoleID.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[str, ...]
            A tuple containing all parent RoleIDs.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
        """
        parents = db.scalars(
            select(HierarchyORM.parent_role_id).where(
                HierarchyORM.child_role_id == role
            )
        ).all()
        if len(parents) == 0 and db.get(RoleORM, role) is None:
            raise PyPermissionError(f"Role '{role}' does not exist!")
        return tuple(parents)

    @classmethod
    def children(cls, *, role: str, db: Session) -> tuple[str, ...]:
        """
        Get all child Roles.

        Parameters
        ----------
        role : str
            The target RoleID.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[str, ...]
            A tuple containing all child RoleIDs.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
        """
        children = db.scalars(
            select(HierarchyORM.child_role_id).where(
                HierarchyORM.parent_role_id == role
            )
        ).all()
        if len(children) == 0 and db.get(RoleORM, role) is None:
            raise PyPermissionError(f"Role '{role}' does not exist!")
        return tuple(children)

    @classmethod
    def ancestors(cls, *, role: str, db: Session) -> tuple[str, ...]:
        """
        Get all ancestor Roles.

        Parameters
        ----------
        role : str
            The target RoleID.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[str, ...]
            A tuple containing all ancestor RoleIDs.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
        """
        root_cte = (
            select(HierarchyORM)
            .where(HierarchyORM.child_role_id == role)
            .cte(name="root_cte", recursive=True)
        )

        traversing_cte = root_cte.alias()
        relations_cte = root_cte.union_all(
            select(HierarchyORM).where(
                HierarchyORM.child_role_id == traversing_cte.c.parent_role_id
            )
        )

        ancestor_relations = (
            db.scalars(select(relations_cte.c.parent_role_id)).unique().all()
        )

        if len(ancestor_relations) == 0 and db.get(RoleORM, role) is None:
            raise PyPermissionError(f"Role '{role}' does not exist!")
        return tuple(ancestor_relations)

    @classmethod
    def descendants(cls, *, role: str, db: Session) -> tuple[str, ...]:
        """
        Get all descending Roles.

        Parameters
        ----------
        role : str
            The target RoleID.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[str, ...]
            A tuple containing all descending RoleIDs.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
        """
        root_cte = (
            select(HierarchyORM)
            .where(HierarchyORM.parent_role_id == role)
            .cte(name="root_cte", recursive=True)
        )

        traversing_cte = root_cte.alias()
        relations_cte = root_cte.union_all(
            select(HierarchyORM).where(
                HierarchyORM.parent_role_id == traversing_cte.c.child_role_id
            )
        )

        descendant_relations = (
            db.scalars(select(relations_cte.c.child_role_id)).unique().all()
        )

        if len(descendant_relations) == 0 and db.get(RoleORM, role) is None:
            raise PyPermissionError(f"Role '{role}' does not exist!")
        return tuple(descendant_relations)

    @classmethod
    def subjects(
        cls, *, role: str, include_descendant_subjects: bool = False, db: Session
    ) -> tuple[str, ...]:
        """
        Get all Subjects assigned to a Role.

        Parameters
        ----------
        role : str
            The target RoleID.
        include_descendant_subjects: bool
            Include all Subjects for descendant Roles.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[str, ...]
            A tuple containing all assigned SubjectIDs.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
        """
        if include_descendant_subjects:
            root_cte = (
                select(RoleORM.id.label("role_id"))
                .where(RoleORM.id == role)
                .cte(recursive=True)
            )

            traversing_cte = root_cte.alias()
            relations_cte = root_cte.union_all(
                select(HierarchyORM.child_role_id).where(
                    HierarchyORM.parent_role_id == traversing_cte.c.role_id
                )
            )
            subjects = (
                db.scalars(
                    select(MemberORM.subject_id).join(
                        relations_cte, MemberORM.role_id == relations_cte.c.role_id
                    )
                )
                .unique()
                .all()
            )
        else:
            subjects = db.scalars(
                select(MemberORM.subject_id).where(MemberORM.role_id == role)
            ).all()

        if len(subjects) == 0 and db.get(RoleORM, role) is None:
            raise PyPermissionError(f"Role '{role}' does not exist!")
        return tuple(subjects)

    @classmethod
    def grant_permission(
        cls,
        *,
        role: str,
        permission: Permission,
        db: Session,
    ) -> None:
        """
        Grant a Permission to a Role.

        Parameters
        ----------
        role : str
            The target RoleID.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
            If the Permission was granted before. TODO
        """
        try:
            policy_orm = PolicyORM(
                role_id=role,
                resource_type=permission.resource_type,
                resource_id=permission.resource_id,
                action=permission.action,
            )
            db.add(policy_orm)
            db.flush()
        except IntegrityError as err:
            db.rollback()
            process_policy_integrity_error(err=err, role=role, permission=permission)

    @classmethod
    def revoke_permission(
        cls,
        *,
        role: str,
        permission: Permission,
        db: Session,
    ) -> None:
        """
        Revoke a Permission from a Role.

        Parameters
        ----------
        role : str
            The target Role ID.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
            If the Permission was not granted before. TODO
        """
        policy_tuple = (
            role,
            permission.resource_type,
            permission.resource_id,
            permission.action,
        )
        policy_orm = db.get(
            PolicyORM,
            policy_tuple,
        )
        if policy_orm is None:
            role_orm = db.get(RoleORM, role)
            if role_orm:
                raise PyPermissionError(
                    f"Permission '{str(permission)}' does not exist!"
                )
            raise PyPermissionError(f"Role '{role}' does not exist!")

        db.delete(policy_orm)
        db.flush()

    @classmethod
    def check_permission(
        cls,
        *,
        role: str,
        permission: Permission,
        db: Session,
    ) -> bool:
        """
        Check if a Role has a Permission.

        Parameters
        ----------
        role : str
            The target RoleID.
        permission : Permission
            The Permission to check for.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        bool
            True if the Permission is granted.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
        """
        root_cte = (
            select(RoleORM.id.label("role_id"))
            .where(RoleORM.id == role)
            .cte(recursive=True)
        )

        traversing_cte = root_cte.alias()
        relations_cte = root_cte.union_all(
            select(HierarchyORM.parent_role_id).where(
                HierarchyORM.child_role_id == traversing_cte.c.role_id
            )
        )
        policy_orms = db.scalars(
            select(PolicyORM)
            .join(relations_cte, PolicyORM.role_id == relations_cte.c.role_id)
            .where(
                PolicyORM.resource_type == permission.resource_type,
                PolicyORM.resource_id.in_((permission.resource_id, "*")),
                PolicyORM.action == permission.action,
            )
        ).all()

        if len(policy_orms) == 0:
            role_orm = db.get(RoleORM, role)
            if role_orm is None:
                raise PyPermissionError(f"Role '{role}' does not exist!")
            return False
        return True

    @classmethod
    def assert_permission(
        cls,
        *,
        role: str,
        permission: Permission,
        db: Session,
    ) -> None:
        """
        Check if a Role has a Permission.

        Parameters
        ----------
        role : str
            The target RoleID.
        permission : Permission
            The Permission to check for.
        db : Session
            The SQLAlchemy session.

        Raises
        ------
        PyPermissionNotGrantedError
            If the Permission is not granted.
        PyPermissionError
            If the target Role does not exist.
        """
        if not cls.check_permission(role=role, permission=permission, db=db):
            raise PermissionNotGrantedError(
                f"Permission '{permission}' is not granted for Role '{role}'!"
            )

    @classmethod
    def permissions(
        cls,
        *,
        role: str,
        inherited: bool = True,
        db: Session,
    ) -> tuple[Permission, ...]:
        """
        Get all granted Permissions for a Role.

        Parameters
        ----------
        role : str
            The target RoleID.
        inherited : bool
            Includes all Permissions inherited by ancestor Roles.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[Permission, ...]
            A tuple containing all granted Permissions.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
        """
        policy_orms = _get_policy_orms_for_role(role=role, inherited=inherited, db=db)
        if len(policy_orms) == 0:
            role_orm = db.get(RoleORM, role)
            if role_orm is None:
                raise PyPermissionError(f"Role '{role}' does not exist!")

        return tuple(
            Permission(
                resource_type=policy_orm.resource_type,
                resource_id=policy_orm.resource_id,
                action=policy_orm.action,
            )
            for policy_orm in policy_orms
        )

    @classmethod
    def policies(
        cls,
        *,
        role: str,
        inherited: bool = True,
        db: Session,
    ) -> tuple[Policy, ...]:
        """
        Get all granted Policies for a Role.

        Parameters
        ----------
        role : str
            The target RoleID.
        inherited : bool
            Includes all Policies inherited by ancestor Roles.
        db : Session
            The SQLAlchemy session.

        Returns
        -------
        tuple[Policies, ...]
            A tuple containing all granted Policies.

        Raises
        ------
        PyPermissionError
            If the target Role does not exist.
        """
        policy_orms = _get_policy_orms_for_role(role=role, inherited=inherited, db=db)

        if len(policy_orms) == 0:
            role_orm = db.get(RoleORM, role)
            if role_orm is None:
                raise PyPermissionError(f"Role '{role}' does not exist!")

        return tuple(
            Policy(
                role=policy_orm.role_id,
                permission=Permission(
                    resource_type=policy_orm.resource_type,
                    resource_id=policy_orm.resource_id,
                    action=policy_orm.action,
                ),
            )
            for policy_orm in policy_orms
        )

    @classmethod
    def actions_on_resource(
        cls,
        *,
        role: str,
        resource_type: str,
        resource_id: str,
        inherited: bool = True,
        db: Session,
    ) -> tuple[str, ...]:
        if inherited:
            root_cte = (
                select(RoleORM.id.label("role_id"))
                .where(RoleORM.id == role)
                .cte(recursive=True)
            )
            traversing_cte = root_cte.alias()
            relations_cte = root_cte.union_all(
                select(HierarchyORM.parent_role_id).where(
                    HierarchyORM.child_role_id == traversing_cte.c.role_id
                )
            )
            selection = (
                select(PolicyORM.action)
                .join(
                    relations_cte,
                    PolicyORM.role_id == relations_cte.c.role_id,
                )
                .where(
                    PolicyORM.resource_type == resource_type,
                    PolicyORM.resource_id.in_((resource_id, "*")),
                )
            )
        else:
            selection = select(PolicyORM.action).where(
                PolicyORM.role_id == role,
                PolicyORM.resource_type == resource_type,
                PolicyORM.resource_id.in_((resource_id, "*")),
            )
        result = db.scalars(selection).all()

        if len(result) == 0:
            role_orm = db.get(RoleORM, role)
            if role_orm is None:
                raise PyPermissionError(f"Role '{role}' does not exist!")
        return tuple(result)


################################################################################
#### Util
################################################################################


def _get_policy_orms_for_role(
    *, role: str, inherited: bool = True, db: Session
) -> Sequence[PolicyORM]:
    # TODO raise IntegrityError if role is unknown and if possible via ORM
    if inherited:
        root_cte = (
            select(RoleORM.id.label("role_id"))
            .where(RoleORM.id == role)
            .cte(recursive=True)
        )

        traversing_cte = root_cte.alias()
        relations_cte = root_cte.union_all(
            select(HierarchyORM.parent_role_id).where(
                HierarchyORM.child_role_id == traversing_cte.c.role_id
            )
        )
        policy_orms = (
            db.scalars(
                select(PolicyORM).join(
                    relations_cte, PolicyORM.role_id == relations_cte.c.role_id
                )
            )
            .unique()
            .all()
        )
    else:
        policy_orms = (
            db.scalars(select(PolicyORM).where(PolicyORM.role_id == role))
            .unique()
            .all()
        )

    return policy_orms
